/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2007 NoMachine, http://www.nomachine.com.          */
/*                                                                        */
/* NXAGENT, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of NoMachine S.r.l.                    */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef __Icons_H__
#define __Icons_H__

#define Pixel XpmPixel
#include <X11/xpm.h>
#undef Pixel

#define NXAGENT_ICON_NAME  "nxagent.xpm"

#define NXAGENT_PLACEHOLDER_NAME  "nxmissing.xpm"

#endif /* __Icons_H__ */
